from construct import IfThenElse, Int32ul, Int64ul, Struct, this

from .common import MINIDUMP_LOCATION_DESCRIPTOR, MINIDUMP_LOCATION_DESCRIPTOR_64

MINIDUMP_MEMORY_DESCRIPTOR = Struct(
    "StartOfMemoryRange" / Int64ul,
    "MemoryLocation"
    / IfThenElse(
        this.StartOfMemoryRange < 0x100000000,
        MINIDUMP_LOCATION_DESCRIPTOR,
        MINIDUMP_LOCATION_DESCRIPTOR_64,
    ),
    "MemoryLocation" / MINIDUMP_LOCATION_DESCRIPTOR,
)

MINIDUMP_MEMORY_LIST = Struct(
    "NumberOfMemoryRanges" / Int32ul,
    # "MemoryRanges"
    # / (MINIDUMP_MEMORY_DESCRIPTOR)[this.NumberOfMemoryRanges]
    # * _track_rva_addr,
)



import logging
import warnings

import rich_click as click
from cryptography.utils import CryptographyDeprecationWarning
from rich import traceback
from rich.logging import RichHandler

warnings.filterwarnings("ignore", category=CryptographyDeprecationWarning)
traceback.install(show_locals=True)

logging.basicConfig(
    level="INFO",
    format="%(message)s",
    datefmt="[%X]",
    handlers=[RichHandler(rich_tracebacks=True, show_path=False, show_time=False)],
)

# ***** RICH CLICK STYLE *****
click.rich_click.MAX_WIDTH = 100
# click.rich_click.USE_RICH_MARKUP = True
click.rich_click.USE_MARKDOWN = True
click.rich_click.SHOW_METAVARS_COLUMN = True
click.rich_click.APPEND_METAVARS_HELP = False

click.rich_click.STYLE_HELPTEXT_FIRST_LINE = "#d24a00"

click.rich_click.STYLE_OPTION = "#f9c300"
click.rich_click.STYLE_OPTIONS_TABLE_BOX = "SIMPLE"
click.rich_click.STYLE_OPTIONS_PANEL_BORDER = "bold #0084a8"
click.rich_click.STYLE_OPTIONS_TABLE_ROW_STYLES = ["#f9c300"]

click.rich_click.STYLE_COMMANDS_TABLE_BOX = "SIMPLE"
click.rich_click.STYLE_COMMANDS_PANEL_BORDER = "bold #0084a8"
click.rich_click.STYLE_COMMANDS_TABLE_ROW_STYLES = ["#f9c300"]

click.rich_click.STYLE_USAGE = "bold #0084a8"
click.rich_click.STYLE_USAGE_COMMAND = "#d24a00 italic"



import construct
from construct import Int32ul, Pointer, Struct, Switch, this

from ..constants import MINIDUMP_STREAM_TYPE
from .common import MINIDUMP_LOCATION_DESCRIPTOR
from .MinidumpMemory64List import MINIDUMP_MEMORY64_LIST
from .MinidumpMemoryList import MINIDUMP_MEMORY_LIST
from .MinidumpThreadList import MINIDUMP_THREAD_LIST

MINIDUMP_DIRECTORY = Struct(
    "StreamType" / construct.Enum(Int32ul, MINIDUMP_STREAM_TYPE),
    "Location" / MINIDUMP_LOCATION_DESCRIPTOR,
    "Data"
    / Pointer(
        this.Location.Rva,
        Switch(
            lambda this: int(this.StreamType),
            {
                MINIDUMP_STREAM_TYPE.MemoryListStream.value: MINIDUMP_MEMORY_LIST,
                MINIDUMP_STREAM_TYPE.Memory64ListStream.value: MINIDUMP_MEMORY64_LIST,
                MINIDUMP_STREAM_TYPE.ThreadListStream.value: MINIDUMP_THREAD_LIST,
            },
            default=None,
        ),
    ),
)



from itertools import accumulate

import construct
from construct import Hex, Int64ul, Pass, Struct, this


def _track_rva_addr(memranges, ctx):
    """Adds the RVA to each range enumerated"""

    # Starting with the BaseRva, we need to keep track of each DataSize to know where the
    # StartOfMemoryRange maps to the RVA in the minidump
    rva_list = list(
        accumulate([memrange.DataSize for memrange in memranges], initial=ctx.BaseRva)
    )
    for index, memrange in enumerate(memranges):
        memrange.DumpRva = rva_list[index]
        # print(memrange.Rva, memrange.DataSize)


MINIDUMP_MEMORY_DESCRIPTOR64 = Struct(
    "StartOfMemoryRange" / Hex(Int64ul),
    "DataSize" / Hex(Int64ul),
    "DumpRva" / Pass,  # We add Pass here as a placeholder for _track_rva_addr
)

MINIDUMP_MEMORY64_LIST = Struct(
    "NumberOfMemoryRanges" / Int64ul,
    "BaseRva" / Hex(Int64ul),
    "MemoryRanges"
    / (MINIDUMP_MEMORY_DESCRIPTOR64)[this.NumberOfMemoryRanges]
    * _track_rva_addr,
)



from construct import Hex, Int32ul, Int64ul, Padding, Pointer, Seek, Struct, this

from .common import (
    MINIDUMP_LOCATION_DESCRIPTOR,
    MINIDUMP_LOCATION_DESCRIPTOR_64,
    UNICODE_STRING,
)
from .MinidumpMemoryList import MINIDUMP_MEMORY_DESCRIPTOR
from .MinidumpMemory64List import MINIDUMP_MEMORY_DESCRIPTOR64

MINIDUMP_PEB = Struct("Padding" / Padding(0x20), "ProcessParameters" / Hex(Int64ul))
MINIDUMP_TEB = Struct("Padding" / Padding(0x60), "PEB" / Hex(Int64ul))

MINIDUMP_PROCESS_PARAMETERS = Struct(
    "Padding" / Padding(0x50),
    "DllPath" / UNICODE_STRING,
    "ImagePathName" / UNICODE_STRING,
    "CommandLine" / UNICODE_STRING,
    "Environment" / Hex(Int64ul),
)

MINIDUMP_THREAD = Struct(
    "ThreadId" / Hex(Int32ul),
    "SuspendCount" / Int32ul,
    "PriorityClass" / Int32ul,
    "Priority" / Int32ul,
    "Teb" / Hex(Int64ul),
    "Stack" / MINIDUMP_MEMORY_DESCRIPTOR64,
    "ThreadContext" / MINIDUMP_LOCATION_DESCRIPTOR,
)

MINIDUMP_THREAD_LIST = Struct(
    "NumberOfThreads" / Int32ul, "Threads" / MINIDUMP_THREAD[this.NumberOfThreads]
)



from construct import Hex, Int16ul, Int32ul, Int64ul, Padding, Seek, Struct

MINIDUMP_LOCATION_DESCRIPTOR = Struct("DataSize" / Int32ul, "Rva" / Hex(Int32ul))
MINIDUMP_LOCATION_DESCRIPTOR_64 = Struct("DataSize" / Int64ul, "Rva" / Hex(Int64ul))

UNICODE_STRING = Struct(
    "Length" / Int16ul,
    "MaximumLength" / Int16ul,
    Padding(4),
    "Buffer" / Hex(Int64ul),
)



YARA_RULES = {
    # Locate x509 structures based on ASN.1
    "x509": {
        "x509": "rule x509 {strings: $a = {30 82 ?? ?? 30 82 ?? ??} condition: $a}",
        "pkcs": "rule pkcs {strings: $a = {30 82 ?? ?? 02 01 00} condition: $a}",
    },
    # Locate RSA structs following "MSRK" magic header
    "symcrypt": {
        "rsa": "rule rsa {strings: $a = {28 00 00 00 4b 52 53 4d ?? ?? ?? ??} condition: $a}",
    },
}



import binascii
from base64 import b64encode
from collections import defaultdict
from pathlib import Path
from struct import unpack

import yara
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.ec import (
    EllipticCurvePrivateKey,
    EllipticCurvePublicKey,
)
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from cryptography.hazmat.primitives.serialization import (
    load_pem_private_key,
    Encoding,
    NoEncryption,
    PrivateFormat,
)
from cryptography.x509 import Certificate, load_der_x509_certificate
from rich import inspect
from rich.console import Console, ConsoleOptions, RenderResult
from rich.table import Table

from ...minidump.minidumpfile import MinidumpFile
from ...minidump.structs.MinidumpMemory64List import MINIDUMP_MEMORY_DESCRIPTOR64
from ..output import get_dumpscan_table
from ..rules import YARA_RULES


class x509Scanner:
    def __init__(self, minidumpfile: MinidumpFile, output: Path) -> None:
        self.rules = yara.compile(sources=YARA_RULES["x509"])
        self.dump = minidumpfile
        self.output = output
        self.matching_objects = defaultdict(list)
        self.modulus_dict = {}
        self.public_private_matches = {}
        self.current_section: MINIDUMP_MEMORY_DESCRIPTOR64 = None

        if output and not output.exists():
            output.mkdir(parents=True)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        table = get_dumpscan_table()
        table.add_column("Rule", style="bold #f9c300")
        table.add_column("Result", style="#008df8")
        table.add_column("Thumbprint", style="#008df8")
        table.add_column("Public Integers (First 20 bytes)")

        for key, values in self.matching_objects.items():
            found_certs = []
            for value in values:
                if isinstance(value, Certificate):
                    thumbprint = (
                        binascii.hexlify(value.fingerprint(hashes.SHA1()))
                        .upper()
                        .decode()
                    )

                    # Clean up output by only printing unique certs
                    if thumbprint in found_certs:
                        continue
                    found_certs.append(thumbprint)

                    public_key = value.public_key()

                    if isinstance(public_key, RSAPublicKey):
                        pubints = format(public_key.public_numbers().n, "x")[
                            :40
                        ].upper()

                    elif isinstance(public_key, EllipticCurvePublicKey):
                        pubints = f"X:{str(public_key.public_numbers().x)[:40]} | Y:{str(public_key.public_numbers().y)[:40]} "

                    table.add_row(
                        key,
                        value.subject.rfc4514_string(),
                        thumbprint,
                        pubints,
                    )
                elif isinstance(value, RSAPrivateKey):
                    result = str(value.key_size)
                    if matching_cert := self.public_private_matches.get(value):
                        result += f"-> {matching_cert.subject.rfc4514_string()}"
                    table.add_row(
                        key,
                        result,
                        None,
                        format(value.private_numbers().public_numbers.n, "x")[
                            :40
                        ].upper(),
                    )
        yield table

    def save_file(self, data: bytes, filename: str):
        if self.output:
            file = self.output / filename
            with file.open("wb") as f:
                f.write(data)

    @classmethod
    def minidump_scan(cls, minidumpfile: MinidumpFile, output: Path) -> "x509Scanner":
        scanner = cls(minidumpfile, output)

        for section, data in minidumpfile.read_all_memory64():
            scanner.current_section = section
            scanner.rules.match(
                data=data,
                callback=scanner.parse_yara_match,
                which_callbacks=yara.CALLBACK_MATCHES,
            )

        scanner.modulus_dict = {}
        for cert in scanner.matching_objects.get("x509"):
            public_key = cert.public_key()

            if isinstance(public_key, RSAPublicKey):
                pub_modulus_str = format(public_key.public_numbers().n, "x").upper()
                scanner.modulus_dict[pub_modulus_str] = cert

        if private_keys := scanner.matching_objects.get("pkcs"):
            for private_key in private_keys:
                priv_modulus_str = format(
                    private_key.private_numbers().public_numbers.n, "x"
                )

                if match := scanner.modulus_dict.get(priv_modulus_str):
                    scanner.public_private_matches[private_key, match]

        return scanner

    def parse_yara_match(self, data):
        rule = data["rule"]
        matching_objects = []

        for match in data["strings"]:
            if obj := self.parse_results(match, rule):
                matching_objects.append(obj)

                if self.output:
                    if rule == "pkcs":
                        output_bytes = obj.private_bytes(
                            Encoding.DER,
                            PrivateFormat.PKCS8,
                            NoEncryption(),
                        )
                        filename = hex(match[0]) + "_" + str(obj.key_size) + ".key"

                    elif rule == "x509":
                        output_bytes = obj.public_bytes(Encoding.DER)
                        thumbprint = (
                            binascii.hexlify(obj.fingerprint(hashes.SHA1()))
                            .upper()
                            .decode()
                        )

                        filename = thumbprint + "_"
                        filename += (
                            obj.subject._attributes[-1]
                            ._attributes[-1]
                            .value.strip('*"/[]:;|,')
                            .split("/")[0]
                        )

                    self.save_file(output_bytes, filename)

        self.matching_objects[rule].extend(matching_objects)
        return yara.CALLBACK_CONTINUE

    def parse_results(self, match: tuple, rule: str):

        # This is the offset from the bytes being scanned
        offset = match[0]

        # Only need first four bytes
        _, cert_size = unpack(">HH", match[2][:4])
        cert_data = self.dump.read_section(self.current_section, offset, cert_size + 4)

        if rule == "x509":
            try:
                return load_der_x509_certificate(cert_data, backend=default_backend())
            except:
                pass
        elif rule == "pkcs":
            pem = (
                b"-----BEGIN RSA PRIVATE KEY-----\n"
                + b64encode(cert_data)
                + b"\n-----END RSA PRIVATE KEY-----"
            )
            try:
                return load_pem_private_key(pem, None, default_backend())
            except:
                pass



from rich import box
from rich.console import Console
from rich.table import Table
from rich.theme import Theme
from rich.style import Style

theme = Theme(
    {
        "repr.attrib_value": "#047bd6",
        "repr.call": Style.null(),
        "repr.ipv6 ": "#008df8",
        "repr.none": "italic #d24a00",
        "repr.number": "#008df8",
        "repr.str": "bright_blue",
        "table.border": "#0084a8",
        "table.cell": "#00a5fa",
        "table.header": "bold italic #d24a00",
    }
)


def get_dumpscan_console():
    return Console(theme=theme)


def get_dumpscan_table():
    return Table(
        expand=False, box=box.SIMPLE_HEAVY, highlight=True, border_style="#047bd6"
    )



import logging
from enum import Enum

from rich import box
from rich.console import Console
from rich.table import Table
from volatility3.cli.text_renderer import QuickTextRenderer
from volatility3.framework import interfaces
from volatility3.framework.renderers import format_hints

from ..common.output import get_dumpscan_console, get_dumpscan_table

log = logging.getLogger("rich")


class RichTableRenderer(QuickTextRenderer):

    name = "richtable"

    def render(self, grid: interfaces.renderers.TreeGrid) -> Table:
        """Renders each column using Rich Table output.
        Args:
            grid: The TreeGrid object to render
        """

        table = get_dumpscan_table()
        [table.add_column(c.name, overflow="fold") for c in grid.columns]

        # This function doesn't need to return anything at all and just updates existing Table object
        def visitor(node: interfaces.renderers.TreeNode, accumulator: None) -> None:
            row = []
            for column_index in range(len(grid.columns)):
                column = grid.columns[column_index]
                renderer = self._type_renderers.get(
                    column.type, self._type_renderers["default"]
                )
                column_rich_text = renderer(node.values[column_index])
                if column_index == 0:
                    column_str = "|" if node.path_depth - 1 else ""
                    row.append(
                        column_str + "-" * (node.path_depth - 1) + str(column_rich_text)
                    )
                else:
                    row.append(str(column_rich_text))
            table.add_row(*row)

        if not grid.populated:
            grid.populate(visitor, None)
        else:
            grid.visit(node=None, function=visitor, initial_accumulator=None)

        return table


class RichTextRenderer(QuickTextRenderer):

    name = "richtext"

    def render(self, grid: interfaces.renderers.TreeGrid) -> str:
        """Renders each column using Rich Table output.
        Args:
            grid: The TreeGrid object to render
        """

        console = get_dumpscan_console()

        line = []
        for column in grid.columns:
            # Ignore the type because namedtuples don't realize they have accessible attributes
            line.append(f"{column.name}")
        console.print("\n{}".format("\t".join(line)))

        def visitor(node: interfaces.renderers.TreeNode, accumulator: Console):
            # Nodes always have a path value, giving them a path_depth of at least 1, we use max just in case
            accumulator.print(
                "*" * max(0, node.path_depth - 1)
                + ("" if (node.path_depth <= 1) else " ")
            )
            line = []
            for column_index in range(len(grid.columns)):
                column = grid.columns[column_index]
                renderer = self._type_renderers.get(
                    column.type, self._type_renderers["default"]
                )
                line.append(renderer(node.values[column_index]))
            accumulator.print("{}".format("\t".join(line)), end="")
            accumulator.file.flush()
            return accumulator

        if not grid.populated:
            grid.populate(visitor, console)
        else:
            grid.visit(node=None, function=visitor, initial_accumulator=console)

        console.print()


class RichRenderOption(str, Enum):
    TABLE = "table"
    TEXT = "text"

    def __init__(self, render: str) -> None:
        if render == "table":
            self.renderer = RichTableRenderer()
        else:
            self.renderer = RichTextRenderer()



import binascii
import warnings
from base64 import b64encode
from struct import unpack
from typing import Callable, Iterable, List, Tuple, Union

import yara
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.types import PRIVATE_KEY_TYPES
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
    load_pem_private_key,
)
from cryptography.utils import CryptographyDeprecationWarning
from cryptography.x509 import Certificate, load_der_x509_certificate
from volatility3.framework import interfaces, objects, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols.windows.extensions import EPROCESS
from volatility3.plugins import yarascan
from volatility3.plugins.windows import pslist, vadyarascan

warnings.filterwarnings("ignore", category=CryptographyDeprecationWarning)


class Dumpcerts(interfaces.plugins.PluginInterface):
    """Dump public and private RSA keys based on ASN-1 structure"""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.PluginRequirement(
                name="vadyarascanner", plugin=vadyarascan.VadYaraScan, version=(1, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                element_type=int,
                description="Process IDs to include (all other processes are excluded)",
                optional=True,
            ),
            requirements.ListRequirement(
                name="name",
                element_type=str,
                description="Process name to include (all other processes are excluded)",
                optional=True,
            ),
            requirements.ChoiceRequirement(
                ["all", "private", "public"],
                name="type",
                default="all",
                description="Types of keys to dump",
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="dump", description="Dump keys", default=False, optional=True
            ),
            requirements.BooleanRequirement(
                name="physical",
                description="Scan physical memory instead of processes",
                default=False,
                optional=True,
            ),
        ]

    def _save_file(self, data: bytes, filename: str, rule: str):
        ext = ".key" if rule == "pkcs" else ".crt"
        with self.open(f"{filename}{ext}") as f:
            f.write(data)

    @classmethod
    def get_yara_rules(cls, key_type: str):
        sources = {}
        if key_type in ["all", "public"]:
            sources[
                "x509"
            ] = "rule x509 {strings: $a = {30 82 ?? ?? 30 82 ?? ??} condition: $a}"

        if key_type in ["all", "private"]:
            sources[
                "pkcs"
            ] = "rule pkcs {strings: $a = {30 82 ?? ?? 02 01 00} condition: $a}"

        return yara.compile(sources=sources)

    @classmethod
    def get_cert_or_pem(
        self,
        layer: interfaces.layers.DataLayerInterface,
        rule_name: str,
        offset: int,
        value: bytes,
    ) -> Union[Certificate, PRIVATE_KEY_TYPES]:
        """Parse the value from the layer and convert to an X509 cert or PEM.

        Args:
            layer (interfaces.layers.DataLayerInterface): Current layer
            rule_name (str): Rule that triggered the match
            offset (int): Offset address
            value (bytes): Bytes representing a certificate or pem

        Returns:
            Union[Certificate, PRIVATE_KEY_TYPES]: Either a Certificate or PEM type
        """
        try:
            _, cert_size = unpack(">HH", value[0:4])
            data = layer.read(offset, cert_size + 4)

            # If x509 triggered, try to create a DER x509 certificate to validate
            if rule_name == "x509":
                rsa_object = load_der_x509_certificate(data, default_backend())

            # If pkcs triggered, try to create a PEM private key to validate
            elif rule_name == "pkcs":
                pem = (
                    b"-----BEGIN RSA PRIVATE KEY-----\n"
                    + b64encode(data)
                    + b"\n-----END RSA PRIVATE KEY-----"
                )
                rsa_object = load_pem_private_key(pem, None, default_backend())

            return rsa_object
        except:
            return None

    @classmethod
    def get_certs_by_process(
        cls,
        context: interfaces.context.ContextInterface,
        proc: interfaces.objects.ObjectInterface,
        key_types: str,
    ) -> Iterable[
        Tuple[
            int,
            interfaces.objects.ObjectInterface,
            str,
            Union[Certificate, PRIVATE_KEY_TYPES],
        ]
    ]:
        """Gets certificates or pem by process

        Args:
            context (interfaces.context.ContextInterface): Context
            proc (interfaces.objects.ObjectInterface): Process object to scan
            key_types (str): Type of key to scan for

        Yields:
            Iterable[ Tuple[ int, interfaces.objects.ObjectInterface, str, Union[Certificate, PRIVATE_KEY_TYPES], ] ]: Scan results
        """
        layer_name = proc.add_process_layer()
        layer = context.layers[layer_name]

        for offset, rule_name, _, value in layer.scan(
            context=context,
            scanner=yarascan.YaraScanner(rules=cls.get_yara_rules(key_types)),
            sections=vadyarascan.VadYaraScan.get_vad_maps(proc),
        ):
            cert_or_pem = cls.get_cert_or_pem(layer, rule_name, offset, value)
            if cert_or_pem:
                yield (offset, proc, rule_name, cert_or_pem)

    @classmethod
    def get_process_certificates(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        symbol_table: str,
        filter_func: Callable[
            [interfaces.objects.ObjectInterface], bool
        ] = lambda _: False,
        key_types: str = "all",
    ) -> Iterable[
        Tuple[
            int,
            interfaces.objects.ObjectInterface,
            str,
            Union[Certificate, PRIVATE_KEY_TYPES],
        ]
    ]:
        """Scans processes for RSA certificates

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            layer_name: The name of the layer on which to operate
            symbol_table: The name of the table containing the kernel symbols
            filter_func: Filter function for listing processes
            key_types: Can be "all", "public", or "private"

        Yields:
            A tuple of offset, EPROCESS, rule name, and certificate or key found by scanning process layer
        """
        for proc in pslist.PsList.list_processes(
            context=context,
            layer_name=layer_name,
            symbol_table=symbol_table,
            filter_func=filter_func,
        ):
            for offset, proc, rule_name, cert_or_pem in cls.get_certs_by_process(
                context, proc, key_types
            ):
                yield (offset, proc, rule_name, cert_or_pem)

    @classmethod
    def get_physical_certificates(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        key_type: str,
    ) -> Iterable[Tuple[int, str, Union[Certificate, PRIVATE_KEY_TYPES],]]:

        layer = context.layers[layer_name]
        for offset, rule_name, _, value in layer.scan(
            context=context,
            scanner=yarascan.YaraScanner(rules=cls.get_yara_rules(key_type)),
        ):
            cert_or_pem = cls.get_cert_or_pem(layer, rule_name, offset, value)
            if cert_or_pem:
                yield (offset, rule_name, cert_or_pem)

    def _generator(self, physical: bool):

        kernel = self.context.modules[self.config["kernel"]]
        pid_list = self.config.get("pid", [])
        name_list = self.config.get("name", [])

        def proc_name_and_pid_filter(proc: EPROCESS):
            try:
                proc_name = proc.ImageFileName.cast(
                    "string",
                    max_length=proc.ImageFileName.vol.count,
                    errors="ignore",
                ).lower()
                return (
                    proc.UniqueProcessId not in pid_list and proc_name not in name_list
                )
            # Specifically, if there is a smear, process info might not be valid
            # So ignore if the process name can't be cast to anything
            except:
                return True

        if pid_list or name_list:
            filter_func = proc_name_and_pid_filter
        else:
            filter_func = lambda x: False

        key_type = self.config.get("type")
        output = self.config.get("dump", False)

        if physical:

            for offset, rule_name, cert_or_pem in self.get_physical_certificates(
                context=self.context, layer_name=kernel.layer_name, key_type=key_type
            ):
                value, output_bytes, thumbprint = self._get_value_and_bytes(
                    rule_name, cert_or_pem
                )

                if output:
                    if rule_name == "pkcs":
                        self._save_file(output_bytes, value, rule_name)
                    elif rule_name == "x509":
                        filename = thumbprint + "_"
                        filename += (
                            cert_or_pem.subject._attributes[-1]
                            ._attributes[-1]
                            .value.strip('*."/[]:;|,')
                            .split("/")[0]
                        )
                        self._save_file(output_bytes, filename, rule_name)

                yield 0, (
                    format_hints.Hex(offset),
                    rule_name,
                    value,
                )
        else:
            for offset, proc, rule_name, rsa_object in self.get_process_certificates(
                context=self.context,
                layer_name=kernel.layer_name,
                symbol_table=kernel.symbol_table_name,
                filter_func=filter_func,
                key_types=key_type,
            ):

                proc_name = proc.ImageFileName.cast(
                    "string",
                    max_length=proc.ImageFileName.vol.count,
                    errors="replace",
                )

                value, output_bytes, thumbprint = self._get_value_and_bytes(
                    rule_name, rsa_object
                )

                if output:
                    if rule_name == "pkcs":
                        self._save_file(
                            output_bytes, f"{hex(offset)}_{value}", rule_name
                        )
                    elif rule_name == "x509":
                        filename = thumbprint + "_"
                        filename += (
                            rsa_object.subject._attributes[-1]
                            ._attributes[-1]
                            .value.strip('*."/\[]:;|,')
                        )

                        self._save_file(output_bytes, filename, rule_name)

                yield 0, (
                    format_hints.Hex(offset),
                    proc.UniqueProcessId,
                    proc_name,
                    rule_name,
                    thumbprint,
                    value,
                )

    def _get_value_and_bytes(
        self, rule_name: str, rsa_object: Union[Certificate, PRIVATE_KEY_TYPES]
    ) -> Tuple[str, bytes, str]:
        """Helper method to get the value and bytes from a Certificate or PEM

        Args:
            rule_name (str): Rule that triggered the match
            rsa_object (Union[Certificate, PRIVATE_KEY_TYPES]): Certificate or PEM

        Returns:
            Tuple[str, bytes]: Value for output, output bytes for saving to file
        """

        # If x509 triggered, value is equal to subject (or thumbprint if subject fails)
        if rule_name == "x509":

            try:
                value = str(rsa_object.subject.rfc4514_string())
                thumbprint = (
                    binascii.hexlify(rsa_object.fingerprint(hashes.SHA1()))
                    .upper()
                    .decode()
                )
            except:
                pass

            output_bytes = rsa_object.public_bytes(Encoding.DER)

        # If pkcs triggered, value is equal to the key size
        elif rule_name == "pkcs":
            value = str(rsa_object.key_size)
            thumbprint = ""
            output_bytes = rsa_object.private_bytes(
                Encoding.DER,
                PrivateFormat.PKCS8,
                NoEncryption(),
            )

        return (value, output_bytes, thumbprint)

    def run(self) -> renderers.TreeGrid:
        physical = self.config.get("physical")
        if physical:
            return renderers.TreeGrid(
                [
                    (f'{"Offset":<8}', format_hints.Hex),
                    ("Rule", str),
                    ("Thumbprint", str),
                    ("Value", str),
                ],
                self._generator(physical),
            )
        else:
            return renderers.TreeGrid(
                [
                    (f'{"Offset":<8}', format_hints.Hex),
                    ("PID", int),
                    (f'{"Process":<8}', str),
                    ("Rule", str),
                    ("Thumbprint/Key Size", str),
                    ("Value", str),
                ],
                self._generator(physical),
            )



from io import RawIOBase
from pathlib import Path

from volatility3.framework.interfaces.plugins import FileHandlerInterface


class DumpscanFileHandler(FileHandlerInterface):
    """Handles writing files to disk"""

    output_dir: Path = Path(".")

    def __init__(self, filename: str) -> None:
        super().__init__(filename)
        if not self.output_dir.exists():
            self.output_dir.mkdir(parents=True)

    def write(self, data: bytes) -> int | None:
        output_file: Path = self.output_dir / self._preferred_filename
        return output_file.write_bytes(data)



from construct import Array, Const, Hex, Int32sl, Int32ul, Int64ul, Struct, Union, this

BCRYPT_RSAKEY = Struct(
    "Length" / Int32ul,
    "Magic" / Const(b"KRSM"),
    "Algid" / Hex(Int32ul),
    "ModBitLen" / Int32ul,
    "Unknown1" / Int32sl,
    "Unknown2" / Int32sl,
    "pAlg" / Hex(Int64ul),
    "pKey" / Hex(Int64ul),
)

SYMCRYPT_RSAKEY = Struct(
    "cbTotalSize" / Int32ul,
    "hasPrivateKey" / Int32ul,
    "nSetBitsOfModulus" / Int32ul,
    "nBitsOfModulus" / Int32ul,
    "nDigitsOfModulus" / Int32ul,
    "nPubExp" / Int32ul,
    "nPrimes" / Int32ul,
    "nBitsOfPrimes" / Array(2, Int32ul),
    "nDigitsOfPrimes" / Array(2, Int32ul),
    "nMaxDigitsOfPrimes" / Array(1, Int32ul),
    "au64PubExp" / Hex(Int64ul),
    "pbPrimes" / Array(2, Hex(Int64ul)),
    "pbCrtInverses" / Array(2, Hex(Int64ul)),
    "pbPrivExps" / Array(1, Hex(Int64ul)),
    "pbCrtPrivExps" / Array(2, Hex(Int64ul)),
    "pmModulus" / Hex(Int64ul),
    "pmPrimes" / Array(2, Hex(Int64ul)),
    "peCrtInverses" / Array(2, Hex(Int64ul)),
    "piPrivExps" / Array(1, Hex(Int64ul)),
    "piCrtPrivExps" / Array(2, Hex(Int64ul)),
    "magic" / Hex(Int64ul),
)

SYMCRYPT_MODULUS_MONTGOMERY = Struct("inv64" / Hex(Int64ul), "rsqr" / Hex(Int32ul))
SYMCRYPT_MODULUS_PSUEDOMERSENNE = Struct("k" / Int32ul)

SYMCRYPT_INT = Struct(
    "type" / Int32ul,
    "nDigits" / Int32ul,
    "cbSize" / Int32ul,
    "magic" / Int64ul,
    "unknown1" / Int64ul,
    "unknown2" / Int32ul,
    "fdef" / Array((this.cbSize - 0x20) // 4, Int32ul),
)

SYMCRYPT_DIVISOR = Struct(
    "type" / Int32ul,
    "nDigits" / Int32ul,
    "cbSize" / Int32ul,
    "nBits" / Int32ul,
    "magic" / Int64ul,
    "td" / Int64ul,
    "int" / SYMCRYPT_INT,
)

SYMCRYPT_MODULUS = Struct(
    "type" / Int32ul,
    "nDigits" / Int32ul,
    "cbSize" / Int32ul,
    "flags" / Int32ul,
    "cbModElement" / Int32ul,
    "magic" / Int64ul,
    "tm"
    / Union(
        0,
        "montgomery" / SYMCRYPT_MODULUS_MONTGOMERY,
        "pseudoMersenne" / SYMCRYPT_MODULUS_PSUEDOMERSENNE,
    ),
    "pUnknown" / Hex(Int64ul),
    "pUnknown2" / Hex(Int64ul),
    "pUnknown3" / Hex(Int64ul),
    "divisor" / SYMCRYPT_DIVISOR,
)



from .common import MINIDUMP_LOCATION_DESCRIPTOR, MINIDUMP_LOCATION_DESCRIPTOR_64
from .MinidumpDirectory import MINIDUMP_DIRECTORY
from .MinidumpHeader import MINIDUMP_HEADER
from .MinidumpMemory64List import MINIDUMP_MEMORY64_LIST, MINIDUMP_MEMORY_DESCRIPTOR64
from .MinidumpMemoryList import MINIDUMP_MEMORY_DESCRIPTOR, MINIDUMP_MEMORY_LIST

__all__ = [
    "MINIDUMP_DIRECTORY",
    "MINIDUMP_HEADER",
    "MINIDUMP_LOCATION_DESCRIPTOR",
    "MINIDUMP_LOCATION_DESCRIPTOR_64",
    "MINIDUMP_MEMORY64_LIST",
    "MINIDUMP_MEMORY_DESCRIPTOR",
    "MINIDUMP_MEMORY_DESCRIPTOR64",
    "MINIDUMP_MEMORY_LIST",
]



import binascii
from pathlib import Path
from struct import unpack
from typing import Dict, Generator, Tuple, TypeVar

import construct
import yara
from construct import RepeatUntil, Seek, Struct, this, CString, Array

from dumpscan.minidump.structs.MinidumpThreadList import (
    MINIDUMP_PEB,
    MINIDUMP_PROCESS_PARAMETERS,
    MINIDUMP_TEB,
)

from ..common.rules import YARA_RULES
from .constants import MINIDUMP_STREAM_TYPE
from .structs import *

T = TypeVar("T")

MINIDUMP_FILE = Struct(
    "Header" / MINIDUMP_HEADER,
    Seek(this.Header.StreamDirectoryRva),
    "Dirs" / construct.Array(this.Header.NumberOfStreams, MINIDUMP_DIRECTORY),
)


class MinidumpFile:
    """Minidump class that parses Windows Minidump format"""

    def __init__(self, filepath: Path):
        self.file = filepath
        self.dump = MINIDUMP_FILE.parse_file(str(filepath.absolute()))
        self._memory_list = self._get_dir(MINIDUMP_STREAM_TYPE.MemoryListStream)
        self._memory64_list = self._get_dir(MINIDUMP_STREAM_TYPE.Memory64ListStream)
        self.thread_list = self._get_dir(MINIDUMP_STREAM_TYPE.ThreadListStream)
        self.memory = self._memory_list or self._memory64_list

    def _get_dir(self, dir_type: T) -> T | None:
        return next(
            (dir for dir in self.dump.Dirs if int(dir.StreamType) == dir_type),
            None,
        )

    def find_section(self, address: int):
        """Finds the section of memory where the address lives"""

        for section in self.memory.Data.MemoryRanges:
            if (
                int(section.StartOfMemoryRange)
                <= address
                <= int(section.StartOfMemoryRange) + int(section.DataSize)
            ):
                return section
        return None

    def read_all_memory64(
        self,
    ) -> Generator[Tuple[Struct, bytes], None, None]:
        """Generator to read all memory64 sections of minidump"""

        with self.file.open("rb") as reader:
            for section in self.memory.Data.MemoryRanges:
                reader.seek(section.DumpRva)
                yield section, reader.read(section.DataSize)

    def read_section(self, section, offset: int, size: int) -> bytes:
        """Reads bytes based on offset from memory section"""

        with self.file.open("rb") as reader:
            reader.seek(section.DumpRva + offset)
            return reader.read(size)

    def read_physical(self, address: int, size: int) -> bytes:
        """Reads bytes based on physical address in dump"""

        if section := self.find_section(address):
            offset = address - section.StartOfMemoryRange
            return self.read_section(section, offset, size)
        else:
            return None

    def get_peb(self):
        """Gets the Process Environment Block (PEB)"""

        thread = self.thread_list.Data.Threads[0]
        teb = MINIDUMP_TEB.parse(self.read_physical(thread.Teb, MINIDUMP_TEB.sizeof()))
        return MINIDUMP_PEB.parse(self.read_physical(teb.PEB, MINIDUMP_PEB.sizeof()))

    def get_process_parameters(self):
        """Gets the process parameter structure in the PEB"""

        peb = self.get_peb()
        return MINIDUMP_PROCESS_PARAMETERS.parse(
            self.read_physical(
                peb.ProcessParameters, MINIDUMP_PROCESS_PARAMETERS.sizeof()
            )
        )

    def get_commandline(self) -> str:
        """Gets the command line strings used to launch the process"""

        params = self.get_process_parameters()
        return self.read_physical(
            params.CommandLine.Buffer, params.CommandLine.Length
        ).decode()


    def get_envars(self) -> Dict[str, str]:
        """Gets the environment variables of the process"""
        params = self.get_process_parameters()

        # Assumption here that the all the envars are in the same section of memory
        section = self.find_section(params.Environment)
        remaining_length = (
            int(section.StartOfMemoryRange) + int(section.DataSize) - params.Environment
        )

        environ_str = self.read_physical(params.Environment, remaining_length)
        envars = RepeatUntil(lambda x, lst, ctx: len(x) == 0, CString("utf16")).parse(environ_str) #fmt: skip
        envar_dict = {}

        # The last value is returned in RepeatUntil and that should be blank so ignore
        for var in envars[:-1]:
            key, value = var.split("=", maxsplit=1)
            envar_dict[key] = value
        return envar_dict


from construct import Const, FlagsEnum, Hex, Int16ul, Int32ul, Struct, Timestamp

from ..constants import MINIDUMP_TYPE

MINIDUMP_HEADER = Struct(
    "Signature" / Const(b"MDMP"),
    "Version" / Int16ul,
    "Implementation" / Int16ul,
    "NumberOfStreams" / Int32ul,
    "StreamDirectoryRva" / Int32ul,
    "Checksum" / Hex(Int32ul),
    "TimeDateStamp" / Timestamp(Int32ul, 1, 1970),
    "Flags" / FlagsEnum(Int32ul, MINIDUMP_TYPE),
)



from pathlib import Path
from typing import List

import typer
from rich import inspect
from rich_click.typer import Typer

from .common.output import get_dumpscan_console, get_dumpscan_table
from .common.scanners.symcrypt import SymcryptScanner
from .common.scanners.x509 import x509Scanner
from .kernel.renderers import RichRenderOption
from .kernel.vol import OS, Volatility
from .minidump.minidumpfile import MinidumpFile

app = Typer(
    name="dumpscan", help="Scan memory dumps for secrets and keys", add_completion=False
)
kernel_app = Typer()
minidump_app = Typer()
app.add_typer(kernel_app, name="kernel", help="Scan kernel dump using volatility")
app.add_typer(minidump_app, name="minidump", help="Scan a user-mode minidump")
console = get_dumpscan_console()


def lowercase_list(value: List[str]) -> List[str]:
    """Callback function to make all strings in list lowercase"""
    return list(map(str.lower, value))


@kernel_app.command(
    name="envar", help="List process environment variables (Only for Windows)"
)
def kernel_envar(
    file: Path = typer.Option(..., "--file", "-f", help="Path to kernel dump"),
    pids: List[int] = typer.Option(
        [], "--pid", "-p", help="Pids to scan. Can be passed multiple times."
    ),
    render: RichRenderOption = typer.Option(
        RichRenderOption.TABLE,
        "--render",
        "-r",
        help="Render output (default: table)",
        case_sensitive=False,
        show_default=False,
    ),
):
    vol = Volatility(file, render, None)
    results = vol.run_envar(list(pids))
    console.print(results)


@kernel_app.command(
    name="cmdline", help="List command line for processes (Only for Windows)"
)
def kernel_cmdline(
    file: Path = typer.Option(..., "--file", "-f", help="Path to kernel dump"),
    pids: List[int] = typer.Option(
        [], "--pid", "-p", help="Pids to scan. Can be passed multiple times."
    ),
    render: RichRenderOption = typer.Option(
        RichRenderOption.TABLE,
        "--render",
        "-r",
        help="Render output (default: table)",
        case_sensitive=False,
        show_default=False,
    ),
):
    vol = Volatility(file, render, None)
    results = vol.run_cmdline(list(pids))
    console.print(results)


@kernel_app.command(
    name="pslist", help="List all the processes and their command line arguments"
)
def kernel_pslist(
    file: Path = typer.Option(..., "--file", "-f", help="Path to kernel dump"),
    os: OS = typer.Option(
        OS.WINDOWS,
        help="Operating system (default: windows)",
        case_sensitive=False,
        show_default=False,
    ),
    render: RichRenderOption = typer.Option(
        RichRenderOption.TABLE,
        "--render",
        "-r",
        help="Render output (default: table)",
        case_sensitive=False,
        show_default=False,
    ),
):
    vol = Volatility(file, render, None)
    results = vol.run_pslist(os)
    console.print(results)


@kernel_app.command(name="x509", help="Scan a kernel-mode dump for x509 certificates")
def kernel_x509(
    file: Path = typer.Option(
        ..., "--file", "-f", help="Path to kernel dump", dir_okay=False
    ),
    output: Path = typer.Option(
        None, "--output", "-o", help="Path to dump objects to disk", file_okay=False
    ),
    pids: List[int] = typer.Option(
        [], "--pid", "-p", help="Pids to scan. Can be passed multiple times."
    ),
    procnames: List[str] = typer.Option(
        [],
        "--procname",
        "-n",
        help="Process names to scan. Can be passed multiple times.",
        callback=lowercase_list,
    ),
    render: RichRenderOption = typer.Option(
        RichRenderOption.TABLE,
        "--render",
        "-r",
        help="Render output (default: table)",
        case_sensitive=False,
        show_default=False,
    ),
):
    vol = Volatility(file, render, output)
    results = vol.run_x509(list(pids), list(procnames), bool(output))
    console.print(results)


@kernel_app.command(
    name="symcrypt", help="Scan a kernel-mode dump for symcrypt objects"
)
def kernel_symcrypt(
    file: Path = typer.Option(
        ..., "--file", "-f", help="Path to kernel dump", dir_okay=False
    ),
    output: Path = typer.Option(
        None, "--output", "-o", help="Path to dump objects to disk", file_okay=False
    ),
    pids: List[int] = typer.Option(
        [], "--pid", "-p", help="Pids to scan. Can be passed multiple times."
    ),
    procnames: List[str] = typer.Option(
        [],
        "--procname",
        "-n",
        help="Process names to scan. Can be passed multiple times.",
        callback=lowercase_list,
    ),
    render: RichRenderOption = typer.Option(
        RichRenderOption.TABLE,
        "--render",
        "-r",
        help="Render output (default: table)",
        case_sensitive=False,
        show_default=False,
    ),
):
    vol = Volatility(file, render, output)
    results = vol.run_symcrypt(list(pids), list(procnames), bool(output))
    console.print(results)


@minidump_app.command(name="x509", help="Scan a minidump for x509 objects")
def minidump_x509(
    file: Path = typer.Option(
        ..., "--file", "-f", help="Path to minidump", dir_okay=False
    ),
    output: Path = typer.Option(
        None, "--output", "-o", help="Path to dump objects to disk", file_okay=False
    ),
):
    minidump_file = MinidumpFile(file.absolute())
    scanner = x509Scanner.minidump_scan(minidump_file, output)
    console.print(scanner)


@minidump_app.command(name="symcrypt", help="Scan a minidump for symcrypt objects")
def minidump_symcrypt(
    file: Path = typer.Option(
        ..., "--file", "-f", help="Path to kernel dump", dir_okay=False
    ),
    output: Path = typer.Option(
        None, "--output", "-o", help="Path to dump objects to disk", file_okay=False
    ),
):

    # Get all the public certs first
    minidump_file = MinidumpFile(file.absolute())
    x509_scanner = x509Scanner.minidump_scan(minidump_file, None)

    # Now look for symcrypt
    symcrypt_scanner = SymcryptScanner.minidump_scan(
        minidump_file, x509_scanner, output
    )
    console.print(symcrypt_scanner)


@minidump_app.command(name="envar", help="Dump the environment variables in a minidump")
def minidump_envar(
    file: Path = typer.Option(
        ..., "--file", "-f", help="Path to kernel dump", dir_okay=False
    ),
):
    # Get all the public certs first
    minidump_file = MinidumpFile(file.absolute())
    # print(minidump_file.dump)

    envars = minidump_file.get_envars()
    table = get_dumpscan_table()
    table.add_column("Variable", style="italic #f9c300")
    table.add_column("Value")
    [table.add_row(k, v) for k, v in envars.items()]
    console.print(table, highlight=False)


@minidump_app.command(name="cmdline", help="Dump the command line string")
def minidump_cmdline(
    file: Path = typer.Option(
        ..., "--file", "-f", help="Path to kernel dump", dir_okay=False
    ),
):
    # Get all the public certs first
    minidump_file = MinidumpFile(file.absolute())
    console.print(
        minidump_file.get_commandline(),
        "\n",
        style="#f9c300",
        highlight=False,
    )


if __name__ == "__main__":
    app()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dumpscan',
 'dumpscan.common',
 'dumpscan.common.scanners',
 'dumpscan.kernel',
 'dumpscan.kernel.plugins',
 'dumpscan.minidump',
 'dumpscan.minidump.structs']

package_data = \
{'': ['*']}

install_requires = \
['arrow>=1.2.2,<2.0.0',
 'capstone>=5.0.0rc2,<6.0.0',
 'cffi>=1.15.0,<2.0.0',
 'construct>=2.10.68,<3.0.0',
 'cryptography>=37.0.2,<38.0.0',
 'jsonschema>=4.5.1,<5.0.0',
 'pefile>=2022.5.30,<2023.0.0',
 'pycryptodome>=3.14.1,<4.0.0',
 'rich-click[typer]>=1.4,<2.0',
 'rich>=12.3.0,<13.0.0',
 'typer>=0.4.1,<0.5.0',
 'yara-python>=3.8.0']

entry_points = \
{'console_scripts': ['dumpscan = dumpscan.main:app']}

setup_kwargs = {
    'name': 'dumpscan',
    'version': '0.1.0',
    'description': 'Scanning memory dumps for secrets using volatility and yara',
    'long_description': '\n<p align="center">\n  <img width="500" height="500" src="https://raw.githubusercontent.com/daddycocoaman/dumpscan/main/docs/dumpscan.png">\n</p>\n\n**Dumpscan** is a command-line tool designed to extract and dump secrets from kernel and Windows Minidump formats. Kernel-dump parsing is provided by [volatility3](https://github.com/volatilityfoundation/volatility3).\n\n## Features\n\n- x509 Public and Private key (PKCS #8/PKCS #1) parsing\n- [SymCrypt](https://github.com/microsoft/SymCrypt) parsing\n  - Supported structures\n    - **SYMCRYPT_RSAKEY** - Determines if the key structure also has a private key\n  - Matching to public certificates found in the same process\n  - More SymCrypt structures to come\n- Environment variables\n- Command line arguments\n\n**Note**: Testing has only been performed on Windows 10 and 11 64-bit hosts and processes. Feel free to file an issue for additional versions. Linux testing TBD.\n\n## Installation\n\nAs a command-line tool, installation is recommended using [pipx](https://github.com/pypa/pipx). This allows for easy updates and well and ensuring it is installed in its own virtual environment.\n\n```\npipx install dumpscan\npipx inject dumpscan git+https://github.com/volatilityfoundation/volatility3#39e812a\n```\n\n## Usage\n\n```\n Usage: dumpscan [OPTIONS] COMMAND [ARGS]...\n\n Scan memory dumps for secrets and keys\n\n╭─ Options ────────────────────────────────────────────────────────────────────────────────────────╮\n│                                                                                                  │\n│  --help         Show this message and exit.                                                      │\n│                                                                                                  │\n╰──────────────────────────────────────────────────────────────────────────────────────────────────╯\n╭─ Commands ───────────────────────────────────────────────────────────────────────────────────────╮\n│                                                                                                  │\n│  kernel     Scan kernel dump using volatility                                                    │\n│  minidump   Scan a user-mode minidump                                                            │\n│                                                                                                  │\n╰──────────────────────────────────────────────────────────────────────────────────────────────────╯\n```\n\nIn the case for subcommands that extract certificates, you can provide `--output/-o <dir>` to output any discovered certificates to disk.  \n\n### Kernel Mode\n\nAs mentioned, kernel analysis is performed by Volatility3. `cmdline`, `envar`, and `pslist` are direct calls to the Volatility3 plugins, while `symcrypt` and `x509` are custom plugins.\n\n```\n Usage: dumpscan kernel [OPTIONS] COMMAND [ARGS]...\n\n Scan kernel dump using volatility\n\n╭─ Options ────────────────────────────────────────────────────────────────────────────────────────╮\n│                                                                                                  │\n│  --help         Show this message and exit.                                                      │\n│                                                                                                  │\n╰──────────────────────────────────────────────────────────────────────────────────────────────────╯\n╭─ Commands ───────────────────────────────────────────────────────────────────────────────────────╮\n│                                                                                                  │\n│  cmdline    List command line for processes (Only for Windows)                                   │\n│  envar      List process environment variables (Only for Windows)                                │\n│  pslist     List all the processes and their command line arguments                              │\n│  symcrypt   Scan a kernel-mode dump for symcrypt objects                                         │\n│  x509       Scan a kernel-mode dump for x509 certificates                                        │\n│                                                                                                  │\n╰──────────────────────────────────────────────────────────────────────────────────────────────────╯\n```\n\n### Minidump Mode\n\nSupports Windows Minidump format.\n\n**Note**: This has only been tested on 64-bit processes on Windows 10+. 32-bit processes requires additional work but isn\'t a priority.\n\n\n```\n Usage: dumpscan minidump [OPTIONS] COMMAND [ARGS]...\n\n Scan a user-mode minidump\n\n╭─ Options ────────────────────────────────────────────────────────────────────────────────────────╮\n│                                                                                                  │\n│  --help         Show this message and exit.                                                      │\n│                                                                                                  │\n╰──────────────────────────────────────────────────────────────────────────────────────────────────╯\n╭─ Commands ───────────────────────────────────────────────────────────────────────────────────────╮\n│                                                                                                  │\n│  cmdline    Dump the command line string                                                         │\n│  envar      Dump the environment variables in a minidump                                         │\n│  symcrypt   Scan a minidump for symcrypt objects                                                 │\n│  x509       Scan a minidump for x509 objects                                                     │\n│                                                                                                  │\n╰──────────────────────────────────────────────────────────────────────────────────────────────────╯\n```\n\n## Built With\n- [volatility3](https://github.com/volatilityfoundation/volatility3)\n- [construct](https://github.com/construct/construct)\n- [yara-python](https://github.com/VirusTotal/yara-python)\n- [typer](https://github.com/tiangolo/typer)\n- [rich](https://github.com/Textualize/rich)\n- [rich_click](https://github.com/ewels/rich-click)\n  \n## Acknowledgements\n- Thanks to [F-Secure](https://github.com/FSecureLABS) and the [physmem2profit](https://github.com/FSecureLABS/physmem2profit) project for providing the idea to use `construct` for parsing minidumps.\n- Thanks to [Skelsec](https://github.com/skelsec) and his [minidump](https://github.com/skelsec/minidump) project which helped me figure out to parse minidumps.\n\n\n## To-Do\n\n- Verify use against 32-bit minidumps\n- Create a coredump parser for Linux process dumps\n- Verify volatility plugins work against Linux kernel dumps\n- Add an HTML report that shows all plugins\n- Code refactoring to make more extensible\n- MORE SECRETS',
    'author': 'Leron Gray',
    'author_email': 'daddycocoaman@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import logging
from enum import Enum
from pathlib import Path
from typing import List

from rich import inspect, print
from volatility3 import framework, plugins
from volatility3.framework import (
    automagic,
    constants,
    contexts,
    interfaces,
    objects,
    renderers,
)
from volatility3.plugins.linux.pslist import PsList as NixPsList
from volatility3.plugins.mac.pslist import PsList as MacPsList
from volatility3.plugins.windows.pslist import PsList as WinPsList

from .filehandler import DumpscanFileHandler
from .plugins.dumpcerts import Dumpcerts
from .plugins.symcrypt import Symcrypt
from .renderers import RichRenderOption

from volatility3.plugins.windows.envars import Envars  # isort: skip
from volatility3.plugins.windows.cmdline import CmdLine

log = logging.getLogger("rich")


class OS(str, Enum):
    WINDOWS = "windows"
    LINUX = "linux"
    MAC = "mac"


PSLIST_PLUGINS = {OS.WINDOWS: WinPsList, OS.LINUX: NixPsList, OS.MAC: MacPsList}


class Volatility:
    def __init__(
        self, dumpfile: Path, renderer: RichRenderOption, output_dir: Path
    ) -> None:
        # Ensure minimum framework version
        framework.require_interface_version(2, 0, 0)

        self.dumpfile = dumpfile
        self.ctx = contexts.Context()
        self.ctx.config[
            "automagic.LayerStacker.single_location"
        ] = self.dumpfile.as_uri()
        self.automagics = automagic.available(self.ctx)

        # Add the plugins path to framework plugins path
        plugins.__path__ = [
            str(Path(__file__).parent / "plugins")
        ] + constants.PLUGINS_PATH
        framework.import_files(plugins, True)

        # Add a file handler
        self.file_handler = DumpscanFileHandler
        if output_dir:
            self.file_handler.output_dir = output_dir

        # Set the render mode
        self.render_mode = renderer

    def _run_plugin(self, plugin: interfaces.plugins.PluginInterface):
        """Runs a plugin"""
        automagics = automagic.choose_automagic(self.automagics, plugin)
        plugin = framework.plugins.construct_plugin(
            self.ctx, automagics, plugin, "plugins", None, self.file_handler
        )
        log.debug("Context config:", self.ctx.config)

        results: renderers.TreeGrid = plugin.run()
        return self.render_mode.renderer.render(results)

    def run_x509(self, pids: List[int], procnames: List[str], dump: bool):
        """Runs dumpcerts plugin on a kernel dump

        Args:
            pids (List[int]): List of pids to filter on
        """

        self.ctx.config["plugins.Dumpcerts.pid"] = pids
        self.ctx.config["plugins.Dumpcerts.name"] = procnames
        self.ctx.config["plugins.Dumpcerts.dump"] = dump

        return self._run_plugin(Dumpcerts)

    def run_symcrypt(self, pids: List[int], procnames: List[str], dump: bool):
        """Runs symcrypt plugin on a kernel dump

        Args:
            pids (List[int]): List of pids to filter on
        """

        self.ctx.config["plugins.Symcrypt.pid"] = pids
        self.ctx.config["plugins.Symcrypt.name"] = procnames
        self.ctx.config["plugins.Symcrypt.dump"] = dump

        return self._run_plugin(Symcrypt)

    def run_pslist(self, os: OS):
        """Runs the pslist plugin for appropriate operating system

        Args:
            os (OS): Operating System
        """
        pslist_plugin = PSLIST_PLUGINS.get(os)
        return self._run_plugin(pslist_plugin)

    def run_envar(self, pids: List[int]):
        """Runs envar plugin on a kernel dump

        Args:
            pids (List[int]): List of pids to filter on
        """
        self.ctx.config["plugins.Envars.pid"] = pids

        return self._run_plugin(Envars)

    def run_cmdline(self, pids: List[int]):
        """Runs cmdline plugin on a kernel dump

        Args:
            pids (List[int]): List of pids to filter on
        """
        self.ctx.config["plugins.CmdLine.pid"] = pids

        return self._run_plugin(CmdLine)



import enum


# Thanks to Skelsec - https://github.com/skelsec/minidump/blob/4945b1011ac202c58003b0198820bc8521eb5af5/minidump/constants.py
class MINIDUMP_STREAM_TYPE(enum.IntEnum):
    """Enum for Minidump stream types"""

    UnusedStream = 0
    ReservedStream0 = 1
    ReservedStream1 = 2
    ThreadListStream = 3
    ModuleListStream = 4
    MemoryListStream = 5
    ExceptionStream = 6
    SystemInfoStream = 7
    ThreadExListStream = 8
    Memory64ListStream = 9
    CommentStreamA = 10
    CommentStreamW = 11
    HandleDataStream = 12
    FunctionTableStream = 13
    UnloadedModuleListStream = 14
    MiscInfoStream = 15
    MemoryInfoListStream = 16
    ThreadInfoListStream = 17
    HandleOperationListStream = 18
    TokenStream = 19
    JavaScriptDataStream = 20
    SystemMemoryInfoStream = 21
    ProcessVmCountersStream = 22
    ThreadNamesStream = 24
    ceStreamNull = 0x8000
    ceStreamSystemInfo = 0x8001
    ceStreamException = 0x8002
    ceStreamModuleList = 0x8003
    ceStreamProcessList = 0x8004
    ceStreamThreadList = 0x8005
    ceStreamThreadContextList = 0x8006
    ceStreamThreadCallStackList = 0x8007
    ceStreamMemoryVirtualList = 0x8008
    ceStreamMemoryPhysicalList = 0x8009
    ceStreamBucketParameters = 0x800A
    ceStreamProcessModuleMap = 0x800B
    ceStreamDiagnosisList = 0x800C
    LastReservedStream = 0xFFFF


# Thanks to Skelsec - https://github.com/skelsec/minidump/blob/4945b1011ac202c58003b0198820bc8521eb5af5/minidump/constants.py
class MINIDUMP_TYPE(enum.IntFlag):
    """Enum for Minidump types"""

    MiniDumpNormal = 0x00000000
    MiniDumpWithDataSegs = 0x00000001
    MiniDumpWithFullMemory = 0x00000002
    MiniDumpWithHandleData = 0x00000004
    MiniDumpFilterMemory = 0x00000008
    MiniDumpScanMemory = 0x00000010
    MiniDumpWithUnloadedModules = 0x00000020
    MiniDumpWithIndirectlyReferencedMemory = 0x00000040
    MiniDumpFilterModulePaths = 0x00000080
    MiniDumpWithProcessThreadData = 0x00000100
    MiniDumpWithPrivateReadWriteMemory = 0x00000200
    MiniDumpWithoutOptionalData = 0x00000400
    MiniDumpWithFullMemoryInfo = 0x00000800
    MiniDumpWithThreadInfo = 0x00001000
    MiniDumpWithCodeSegs = 0x00002000
    MiniDumpWithoutAuxiliaryState = 0x00004000
    MiniDumpWithFullAuxiliaryState = 0x00008000
    MiniDumpWithPrivateWriteCopyMemory = 0x00010000
    MiniDumpIgnoreInaccessibleMemory = 0x00020000
    MiniDumpWithTokenInformation = 0x00040000
    MiniDumpWithModuleHeaders = 0x00080000
    MiniDumpFilterTriage = 0x00100000
    MiniDumpValidTypeFlags = 0x001FFFFF



import binascii
from collections import defaultdict
from dataclasses import asdict, dataclass
from pathlib import Path
from struct import unpack
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.rsa import (
    RSAPrivateNumbers,
    RSAPublicNumbers,
    rsa_crt_dmp1,
    rsa_crt_dmq1,
    rsa_crt_iqmp,
    rsa_recover_prime_factors,
)
from cryptography.hazmat.primitives.serialization.pkcs12 import (
    serialize_key_and_certificates,
)
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
)

import yara
from rich import inspect
from rich.console import Console, ConsoleOptions, RenderResult

from ...common.structs import *
from ...minidump.minidumpfile import MinidumpFile
from ...minidump.structs.MinidumpMemory64List import MINIDUMP_MEMORY_DESCRIPTOR64
from ..output import get_dumpscan_table
from ..rules import YARA_RULES
from .x509 import x509Scanner


@dataclass
class SymcryptRSAResult:
    address: str
    hasPrivateKey: int
    modulus: int
    match: str


class SymcryptScanner:
    def __init__(
        self, minidumpfile: MinidumpFile, x509_scanner: x509Scanner, output: Path
    ) -> None:
        self.rules = yara.compile(sources=YARA_RULES["symcrypt"])
        self.dump = minidumpfile
        self.output = output
        self.x509 = x509_scanner
        self.matching_objects = defaultdict(list)
        self.modulus_dict = {}
        self.public_private_matches = {}
        self.current_section: MINIDUMP_MEMORY_DESCRIPTOR64 = None

        if output and not output.exists():
            output.mkdir(parents=True)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        # RSA key results
        table = get_dumpscan_table()
        table.add_column("Rule", style="bold italic #f9c300")
        table.add_column("Address")
        table.add_column("HasPrivateKey")
        table.add_column("Modulus (First 20 bytes)", style="bold #f9c300")
        table.add_column("Matching Certificate")
        for result in self.matching_objects.get("rsa", []):
            table.add_row("rsa", *map(str, asdict(result).values()))
        yield table

    @classmethod
    def minidump_scan(
        cls, minidumpfile: MinidumpFile, x509_scanner: x509Scanner, output: Path
    ) -> "SymcryptScanner":
        scanner = cls(minidumpfile, x509_scanner, output)

        for section, data in minidumpfile.read_all_memory64():
            scanner.current_section = section
            scanner.rules.match(
                data=data,
                callback=scanner.parse_yara_match,
                which_callbacks=yara.CALLBACK_MATCHES,
            )
        return scanner

    def parse_yara_match(self, data):
        parsing_functions = {"rsa": self._parse_rsakey}

        rule = data["rule"]
        matching_objects = []

        if parse_func := parsing_functions.get(rule):
            for match in data["strings"]:
                if obj := parse_func(match, rule):
                    matching_objects.append(obj)

        self.matching_objects[rule].extend(matching_objects)
        return yara.CALLBACK_CONTINUE

    def _parse_rsakey(self, match: tuple, rule: str):

        # This is the offset from the bytes being scanned
        offset = match[0]
        physical_address = self.current_section.StartOfMemoryRange + offset

        # The expected size of the structure is 0x28 in length
        bcrypt_rsakey = BCRYPT_RSAKEY.parse(
            self.dump.read_section(self.current_section, offset, 0x28)
        )
        key_size = unpack("I", self.dump.read_physical(bcrypt_rsakey.pKey, 4))[0]
        key = SYMCRYPT_RSAKEY.parse(
            self.dump.read_physical(bcrypt_rsakey.pKey, key_size)
        )

        # Get the cbSize of modulus (pmModulus + 8) then parse into Modulus struct
        modulus_size = unpack("I", self.dump.read_physical(key.pmModulus + 8, 4))[0]
        modulus = SYMCRYPT_MODULUS.parse(
            self.dump.read_physical(key.pmModulus, modulus_size)
        )
        # Zfill is important here for alignment
        # Additionally, we have to read the list of integers (def) backwards
        mod_str = "".join(
            [format(i, "x").zfill(8) for i in modulus.divisor.int.fdef[::-1]]
        ).upper()
        matching_cert = None
        matching_cert_value = ""

        if self.x509:
            if matching_cert := self.x509.modulus_dict.get(mod_str):
                thumbprint = (
                    binascii.hexlify(matching_cert.fingerprint(hashes.SHA1()))
                    .upper()
                    .decode()
                )
                subject = matching_cert.subject.rfc4514_string()
                matching_cert_value = f"[green]{thumbprint}[/green] -> {subject}"

                if key.hasPrivateKey and self.output:
                    # We could write extra code to parse each of the primes but we need to get the private exponent (d) anyway
                    # So less work to just pull private exponent and derive the primes from n,e,d
                    private_exp_size = unpack(
                        "I", self.dump.read_physical(key.piPrivExps[0] + 8, 4)
                    )[0]
                    private_exp_modulus = SYMCRYPT_INT.parse(
                        self.dump.read_physical(key.piPrivExps[0], private_exp_size)
                    )

                    private_exp_hexstr = "".join(
                        [
                            format(i, "x").zfill(8)
                            for i in private_exp_modulus.fdef[::-1]
                        ]
                    )

                    # Get p and q from modulus, public exponent, and private exponent
                    d = int(private_exp_hexstr, 16)  # Private exponent
                    n = int(mod_str, 16)  # Modulus
                    e = key.au64PubExp  # Public exponent
                    p, q = rsa_recover_prime_factors(n, e, d)

                    # fmt: on

                    # We need to create the public numbers to pass to the private numbers
                    # All of these numbers exist in the parsed Structs, but easier to call helper functions
                    public_numbers = RSAPublicNumbers(e, n)
                    private_numbers = RSAPrivateNumbers(
                        p=p,
                        q=q,
                        d=d,
                        dmp1=rsa_crt_dmp1(d, p),
                        dmq1=rsa_crt_dmq1(d, q),
                        iqmp=rsa_crt_iqmp(p, q),
                        public_numbers=public_numbers,
                    )
                    private_key = private_numbers.private_key()
                    if matching_cert:
                        pfx = serialize_key_and_certificates(
                            thumbprint.encode(),
                            private_key,
                            matching_cert,
                            None,
                            NoEncryption(),
                        )
                        filename = thumbprint + "_"
                        filename += (
                            matching_cert.subject._attributes[-1]
                            ._attributes[-1]
                            .value.strip('*."/[]:;|,')
                            .split("/")[0]
                        )
                        with open(f"{self.output / filename}.pfx", "wb") as f:
                            f.write(pfx)

        return SymcryptRSAResult(
            hex(physical_address), key.hasPrivateKey, mod_str[:40], matching_cert_value
        )



import logging
import warnings
from struct import unpack
from typing import Dict, List

import yara
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.rsa import (
    RSAPrivateNumbers,
    RSAPublicKey,
    RSAPublicNumbers,
    rsa_crt_dmp1,
    rsa_crt_dmq1,
    rsa_crt_iqmp,
    rsa_recover_prime_factors,
)
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
)
from cryptography.hazmat.primitives.serialization.pkcs12 import (
    serialize_key_and_certificates,
)
from cryptography.utils import CryptographyDeprecationWarning
from cryptography.x509 import Certificate
from rich import inspect
from volatility3.framework import interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols.windows.extensions import EPROCESS
from volatility3.plugins import yarascan
from volatility3.plugins.windows import pslist, vadyarascan

from ...common.structs import *
from .dumpcerts import Dumpcerts

warnings.filterwarnings("ignore", category=CryptographyDeprecationWarning)

log = logging.getLogger("rich")

# fmt: off
def get_yara_rules():
    sources = {}

    # Follow the start of the BCRYPT_RSAKEY Struct
    sources["symcrypt_rsa_key"] = "rule symcrypt_rsa_key {strings: $a = {28 00 00 00 4b 52 53 4d ?? ?? ?? ?? 00 (08 | 0c | 10) 00 00 01} condition: $a}"

    return yara.compile(sources=sources)
# fmt: on


class Symcrypt(interfaces.plugins.PluginInterface):
    """Dump symcrypt keys"""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Kernel Layer",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.PluginRequirement(
                name="vadyarascanner", plugin=vadyarascan.VadYaraScan, version=(1, 0, 0)
            ),
            requirements.PluginRequirement(
                name="dumpcerts", plugin=Dumpcerts, version=(1, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                element_type=int,
                description="Process IDs to include (all other processes are excluded)",
                optional=True,
            ),
            requirements.ListRequirement(
                name="name",
                element_type=str,
                description="Process name to include (all other processes are excluded)",
                optional=True,
            ),
            requirements.ListRequirement(
                name="matches",
                description="List of modulus to compare to start of private key to match",
                default=[],
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="dump",
                description="Dump PFX when a match is found",
                default=False,
                optional=True,
            ),
        ]

    def _generator(self):

        kernel = self.context.modules[self.config["kernel"]]
        pid_list = self.config.get("pid", [])
        name_list = self.config.get("name", [])

        def proc_name_and_pid_filter(proc: EPROCESS):
            if not pid_list and not name_list:
                return False

            if proc.is_valid() and proc.UniqueProcessId:
                try:
                    proc_name = proc.ImageFileName.cast(
                        "string",
                        max_length=proc.ImageFileName.vol.count,
                        errors="ignore",
                    ).lower()
                    return (
                        proc.UniqueProcessId not in pid_list
                        and proc_name not in name_list
                    )
                # Specifically, if there is a smear, process info might not be valid
                # So ignore if the process name can't be cast to anything
                except:
                    return True
            else:
                return True

        filter_func = proc_name_and_pid_filter
        output = self.config.get("dump")
        modulus_matches = self.config.get("matches")
        # Get all the processes
        for proc in pslist.PsList.list_processes(
            context=self.context,
            layer_name=kernel.layer_name,
            symbol_table=kernel.symbol_table_name,
            filter_func=filter_func,
        ):
            # If process can't be added, just ignore it
            try:
                layer_name = proc.add_process_layer()
            except:
                continue

            layer = self.context.layers[layer_name]
            proc_name = proc.ImageFileName.cast(
                "string",
                max_length=proc.ImageFileName.vol.count,
                errors="replace",
            )

            # This will hold all of the public certificates to match against if a hit occurs
            public_certs: Dict[str, Certificate] = {}

            for offset, rule_name, _, _ in layer.scan(
                context=self.context,
                scanner=yarascan.YaraScanner(rules=get_yara_rules()),
                sections=vadyarascan.VadYaraScan.get_vad_maps(proc),
            ):
                # If there is a match for key, then...
                # Get all of the public certs in the process
                # Keep track of modulus and cert object
                if not public_certs:
                    for _, _, _, cert in Dumpcerts.get_certs_by_process(
                        context=self.context,
                        proc=proc,
                        key_types="public",
                    ):
                        try:
                            # If an RSA key, grab the modulus and convert it to a string
                            if isinstance(cert.public_key(), RSAPublicKey):
                                public_certs[
                                    format(cert.public_key().public_numbers().n, "x")
                                ] = cert
                        except:
                            continue

                # Make the bcrypt_RSA_KEY structure to get the pointer to key
                bcrypt_rsakey = BCRYPT_RSAKEY.parse(layer.read(offset, 0x28))

                # Parse the key into a SYMCRYPT_RSA_KEY struct
                # If we can't, just move on
                try:
                    key_total_size = unpack("I", layer.read(bcrypt_rsakey.pKey, 4))[0]
                    key = SYMCRYPT_RSAKEY.parse(
                        layer.read(bcrypt_rsakey.pKey, key_total_size)
                    )
                except Exception as e:
                    yield 0, (
                        format_hints.Hex(offset),
                        proc.UniqueProcessId,
                        proc_name,
                        rule_name,
                        -1,
                        f"Corrupt - {e}",
                    )
                    continue

                # Get the cbSize of modulus (pmModulus + 8) then parse into Modulus struct
                modulus_size = unpack("I", layer.read(key.pmModulus + 8, 4))[0]
                modulus = SYMCRYPT_MODULUS.parse(
                    layer.read(key.pmModulus, modulus_size)
                )

                # Zfill is important here for alignment
                # Additionally, we have to read the list of integers (def) backwards
                mod_str = "".join(
                    [format(i, "x").zfill(8) for i in modulus.divisor.int.fdef[::-1]]
                )

                # Look for a matching cert in the process. If one is found, print thumbprint and subject
                match_string = ""

                matching_cert = public_certs.get(mod_str, None)
                if matching_cert:
                    thumbprint = "".join(
                        "{:02X}".format(b)
                        for b in matching_cert.fingerprint(hashes.SHA1())
                    )

                    subject = matching_cert.subject.rfc4514_string()
                    match_string = f"[green]{thumbprint}[/green] -> {subject}"
                else:
                    # If there's no matching cert, then print out first 20 bytes of modulus
                    # Check to see if modulus matches what user requests
                    if modulus_matches:
                        for modulus in modulus_matches:
                            if mod_str.upper().startswith(modulus.upper()):
                                match_string = f"[green]{modulus}[/]"

                # If there is a matching cert and there is a private key, make PFX or print pem!
                if key.hasPrivateKey and output:
                    # fmt: off

                    # We could write extra code to parse each of the primes but we need to get the private exponent (d) anyway
                    # So less work to just pull private exponent and derive the primes from n,e,d
                    private_exp_size = unpack("I", layer.read(key.piPrivExps[0] + 8, 4))[0]
                    private_exp_modulus = SYMCRYPT_INT.parse(
                        layer.read(key.piPrivExps[0], private_exp_size)
                    )
                    private_exp_hexstr = "".join(
                        [format(i, "x").zfill(8) for i in private_exp_modulus.fdef[::-1]]
                    )

                    # Get p and q from modulus, public exponent, and private exponent
                    d = int(private_exp_hexstr, 16) # Private exponent
                    n = int(mod_str, 16)            # Modulus
                    e = key.au64PubExp              # Public exponent
                    p, q = rsa_recover_prime_factors(n, e, d)

                    # fmt: on

                    # We need to create the public numbers to pass to the private numbers
                    # All of these numbers exist in the parsed Structs, but easier to call helper functions
                    public_numbers = RSAPublicNumbers(e, n)
                    private_numbers = RSAPrivateNumbers(
                        p=p,
                        q=q,
                        d=d,
                        dmp1=rsa_crt_dmp1(d, p),
                        dmq1=rsa_crt_dmq1(d, q),
                        iqmp=rsa_crt_iqmp(p, q),
                        public_numbers=public_numbers,
                    )
                    private_key = private_numbers.private_key()
                    if matching_cert:
                        pfx = serialize_key_and_certificates(
                            thumbprint.encode(),
                            private_key,
                            matching_cert,
                            None,
                            NoEncryption(),
                        )
                        filename = thumbprint + "_"
                        filename += (
                            matching_cert.subject._attributes[-1]
                            ._attributes[-1]
                            .value.strip('*."/[]:;|,')
                            .split("/")[0]
                        )
                        with self.open(f"{filename}.pfx") as f:
                            f.write(pfx)
                    else:
                        with self.open(f"{mod_str[:40].upper()}_modulus.key") as f:
                            f.write(
                                private_key.private_bytes(
                                    encoding=Encoding.DER,
                                    format=PrivateFormat.PKCS8,
                                    encryption_algorithm=NoEncryption(),
                                )
                            )

                yield 0, (
                    format_hints.Hex(offset),
                    proc.UniqueProcessId,
                    f"[#f9c300]{proc_name}[/]",
                    f"[#f9c300]{rule_name}[/]",
                    key.hasPrivateKey,
                    f"[#f9c300]{mod_str[:40].upper()}[/]",
                    match_string,
                )

    def run(self) -> renderers.TreeGrid:
        return renderers.TreeGrid(
            [
                (f'{"Offset":<8}', format_hints.Hex),
                ("PID", int),
                (f'{"Process":<8}', str),
                ("Rule", str),
                ("HasPrivateKey", int),
                ("Modulus (First 20 Bytes)", str),
                ("Matching", str),
            ],
            self._generator(),
        )
